package eel6788.agents.directeddiffusion;

import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;

import yaes.framework.world.map.Location;
import yaes.framework.world.sensornetwork.Perception;
import yaes.ui.text.TextUi;

/*
 * This class manages both the gradients - which are directions towards nodes
 * which are interested - as well as records incoming observations 
 */
public class Interest {
	private final String interestType;
	private int smallestInterval = Integer.MAX_VALUE;
	private int maxDuration = 0;
	private final Rectangle2D.Double rectangle;
	private final HashMap<String, Gradient> gradients =
			new HashMap<String, Gradient>();
	private final Map<Integer, List<String>> processedPerceptions =
			new TreeMap<Integer, List<String>>();

	/**
	 * Creates an interest with the given observation type, and it creates the
	 * first gradient as well.
	 * 
	 * @param interestType
	 */
	public Interest(String interestType, String sensorName, int interval,
			int duration, Rectangle2D.Double rectangle) {
		this.interestType = interestType;
		gradients.put(sensorName, new Gradient(sensorName, interval, duration));
		maxDuration = duration;
		smallestInterval = interval;
		this.rectangle = rectangle;
	}
	
	/**
	 * Creates an interest without gradients, to be used in the sink
	 */
	public Interest(String interestType, Rectangle2D.Double rectangle) {
		this.interestType = interestType;
		this.rectangle = rectangle;
	}

	/**
	 * Updates the gradients based on an incoming message.
	 * 
	 * @param sensorName
	 * @param interval
	 * @param duration
	 * @return if the interest was changed
	 */
	public boolean updateInterestOnNewMessage(String sensorName, int interval,
			int duration) {
		boolean retval = false;
		Gradient theGradient = gradients.get(sensorName);
		if (theGradient == null) {
			theGradient = new Gradient(sensorName, interval, duration);
			gradients.put(sensorName, theGradient);
			retval = true;
		}
		theGradient.update(interval, duration);
		if (interval < smallestInterval) {
			smallestInterval = interval;
			retval = true;
		}
		if (duration > maxDuration) {
			maxDuration = duration;
			retval = true;
		}
		return retval;
	}

	/**
	 * Discards the expired gradients
	 */
	public void updateInterestBasedOnTime(int time) {
		final List<String> toRemove = new ArrayList<String>();
		for (final Gradient gradient : gradients.values()) {
			if (gradient.getDuration() < time) {
				toRemove.add(gradient.getSensorName());
			}
		}
		for (final String sensorName : toRemove) {
			gradients.remove(sensorName);
		}
		// if anything was removed recalculate maxDuration and smallestInterval
		if (toRemove.size() != 0) {
			maxDuration = 0;
			smallestInterval = Integer.MAX_VALUE;
			for (final Gradient gradient : gradients.values()) {
				if (maxDuration < gradient.getDuration()) {
					maxDuration = gradient.getDuration();
				}
				if (smallestInterval > gradient.getInterval()) {
					smallestInterval = gradient.getInterval();
				}
			}
		}
	}

	/**
	 * Returns an unmodifiable collection of the gradients
	 * 
	 * @return
	 */
	public Collection<Gradient> getGradientCollection() {
		return Collections.unmodifiableCollection(gradients.values());
	}

	/**
	 * Returns true if the location is of interest for this interest.
	 * 
	 * @param location
	 * @return
	 */
	public boolean isLocationOfInterest(Location location) {
		final Point2D.Double loc = location.asPoint();
		return rectangle.contains(loc);
	}

	/**
	 * Prints out the interest for debugging purposes
	 */
	@Override
	public String toString() {
		final StringBuffer buffer =
				new StringBuffer("Interest: " + interestType);
		buffer.append("  maxDuration: " + maxDuration + " smallestInterval: "
				+ smallestInterval);
		for (final Gradient gradient : gradients.values()) {
			buffer.append("\n"
					+ TextUi.shiftTextWithSpaces(3, gradient.toString()));
		}
		for (Integer perception : processedPerceptions.keySet()) {
			List<String> reporters = processedPerceptions.get(perception);
			buffer.append("\n"
					+ TextUi.shiftTextWithSpaces(5, "Perception: " + perception
							+ " reported by: "));
			for (String reporter : reporters) {
				buffer.append(reporter + " ");
			}
		}
		return buffer.toString();
	}

	public String getInterestType() {
		return interestType;
	}

	/**
	 * Verifies whether this perception was previously processed
	 * 
	 * @param perceptionId
	 * @return true if the perception has been previously seen
	 */
	public boolean processPerceptionId(int perceptionId, String sender) {
		List<String> previousReports = processedPerceptions.get(perceptionId);
		if (previousReports != null) {
			previousReports.add(sender);
			return true;
		}
		previousReports = new ArrayList<String>();
		previousReports.add(sender);
		processedPerceptions.put(perceptionId, previousReports);
		return false;
	}

	public Rectangle2D.Double getRectangle() {
		return rectangle;
	}

	public Map<Integer, List<String>> getProcessedPerceptions() {
		return processedPerceptions;
	}
}
